# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import Callable, List

import pytest
import sciline
import scipp as sc

from ess import isissans as isis
from ess import sans
from ess.isissans import MonitorOffset, SampleOffset, sans2d
from ess.sans.types import (
    BackgroundRun,
    BackgroundSubtractedIofQ,
    BeamCenter,
    CorrectForGravity,
    DimsToKeep,
    DirectBeam,
    DirectBeamFilename,
    EmptyBeamRun,
    Filename,
    Incident,
    IofQ,
    MaskedData,
    NeXusMonitorName,
    NonBackgroundWavelengthRange,
    QBins,
    RawData,
    ReturnEvents,
    SampleRun,
    SolidAngle,
    Transmission,
    UncertaintyBroadcastMode,
    WavelengthBands,
    WavelengthBins,
    WavelengthMask,
)


def make_params() -> dict:
    params = isis.default_parameters()
    params[WavelengthBins] = sc.linspace(
        'wavelength', start=2.0, stop=16.0, num=141, unit='angstrom'
    )
    params[WavelengthMask] = sc.DataArray(
        data=sc.array(dims=['wavelength'], values=[True]),
        coords={
            'wavelength': sc.array(
                dims=['wavelength'], values=[2.21, 2.59], unit='angstrom'
            )
        },
    )
    params[sans2d.LowCountThreshold] = sc.scalar(100.0, unit='counts')

    params[QBins] = sc.linspace(
        dim='Q', start=0.01, stop=0.55, num=141, unit='1/angstrom'
    )
    params[DirectBeamFilename] = isis.data.sans2d_tutorial_direct_beam()
    params[Filename[SampleRun]] = isis.data.sans2d_tutorial_sample_run()
    params[Filename[BackgroundRun]] = isis.data.sans2d_tutorial_background_run()
    params[Filename[EmptyBeamRun]] = isis.data.sans2d_tutorial_empty_beam_run()

    params[NeXusMonitorName[Incident]] = 'monitor2'
    params[NeXusMonitorName[Transmission]] = 'monitor4'
    params[SampleOffset] = sc.vector([0.0, 0.0, 0.053], unit='m')
    params[MonitorOffset[Transmission]] = sc.vector([0.0, 0.0, -6.719], unit='m')

    params[NonBackgroundWavelengthRange] = sc.array(
        dims=['wavelength'], values=[0.7, 17.1], unit='angstrom'
    )
    params[CorrectForGravity] = True
    params[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.upper_bound
    params[ReturnEvents] = False
    params[DimsToKeep] = tuple()

    return params


def sans2d_providers():
    return list(
        sans.providers
        + isis.providers
        + isis.sans2d.providers
        + isis.mantidio.providers
        + (
            isis.data.load_tutorial_direct_beam,
            isis.data.load_tutorial_run,
            isis.data.transmission_from_background_run,
            isis.data.transmission_from_sample_run,
            sans.beam_center_finder.beam_center_from_center_of_mass,
        )
    )


def test_can_create_pipeline():
    sciline.Pipeline(sans2d_providers(), params=make_params())


@pytest.mark.parametrize(
    'uncertainties',
    [UncertaintyBroadcastMode.drop, UncertaintyBroadcastMode.upper_bound],
)
def test_pipeline_can_compute_background_subtracted_IofQ(uncertainties):
    params = make_params()
    params[UncertaintyBroadcastMode] = uncertainties
    pipeline = sciline.Pipeline(sans2d_providers(), params=params)
    result = pipeline.compute(BackgroundSubtractedIofQ)
    assert result.dims == ('Q',)


def test_pipeline_can_compute_background_subtracted_IofQ_in_wavelength_bands():
    params = make_params()
    params[WavelengthBands] = sc.linspace(
        'wavelength', start=2.0, stop=16.0, num=11, unit='angstrom'
    )
    pipeline = sciline.Pipeline(sans2d_providers(), params=params)
    result = pipeline.compute(BackgroundSubtractedIofQ)
    assert result.dims == ('band', 'Q')
    assert result.sizes['band'] == 10


def test_pipeline_wavelength_bands_is_optional():
    params = make_params()
    pipeline = sciline.Pipeline(sans2d_providers(), params=params)
    noband = pipeline.compute(BackgroundSubtractedIofQ)
    assert pipeline.compute(WavelengthBands) is None
    band = sc.linspace('wavelength', 2.0, 16.0, num=2, unit='angstrom')
    pipeline[WavelengthBands] = band
    assert sc.identical(band, pipeline.compute(WavelengthBands))
    withband = pipeline.compute(BackgroundSubtractedIofQ)
    assert sc.identical(noband, withband)


def test_workflow_is_deterministic():
    params = make_params()
    params[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.drop
    pipeline = sciline.Pipeline(sans2d_providers(), params=params)
    # This is Sciline's default scheduler, but we want to be explicit here
    scheduler = sciline.scheduler.DaskScheduler()
    graph = pipeline.get(IofQ[SampleRun], scheduler=scheduler)
    reference = graph.compute().data
    result = graph.compute().data
    assert sc.identical(sc.values(result), sc.values(reference))


def test_pipeline_raises_VariancesError_if_normalization_errors_not_dropped():
    params = make_params()
    params[
        NonBackgroundWavelengthRange
    ] = None  # Make sure we raise in iofq_denominator
    params[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.fail
    pipeline = sciline.Pipeline(sans2d_providers(), params=params)
    with pytest.raises(sc.VariancesError):
        pipeline.compute(BackgroundSubtractedIofQ)


def test_uncertainty_broadcast_mode_drop_yields_smaller_variances():
    params = make_params()
    # Errors with the full range have some NaNs or infs
    params[QBins] = sc.linspace(
        dim='Q', start=0.01, stop=0.5, num=141, unit='1/angstrom'
    )
    params[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.drop
    pipeline = sciline.Pipeline(sans2d_providers(), params=params)
    drop = pipeline.compute(IofQ[SampleRun]).data
    params[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.upper_bound
    pipeline = sciline.Pipeline(sans2d_providers(), params=params)
    upper_bound = pipeline.compute(IofQ[SampleRun]).data
    assert sc.all(sc.variances(drop) < sc.variances(upper_bound)).value


def test_pipeline_can_visualize_background_subtracted_IofQ():
    pipeline = sciline.Pipeline(sans2d_providers(), params=make_params())
    pipeline.visualize(BackgroundSubtractedIofQ)


def test_pipeline_can_compute_intermediate_results():
    pipeline = sciline.Pipeline(sans2d_providers(), params=make_params())
    result = pipeline.compute(SolidAngle[SampleRun])
    assert result.dims == ('spectrum',)


# TODO See scipp/sciline#57 for plans on a builtin way to do this
def as_dict(funcs: List[Callable[..., type]]) -> dict:
    from typing import get_type_hints

    return dict(zip([get_type_hints(func)['return'] for func in funcs], funcs))


def pixel_dependent_direct_beam(
    filename: DirectBeamFilename, shape: RawData[SampleRun]
) -> DirectBeam:
    direct_beam = isis.data.load_tutorial_direct_beam(filename)
    sizes = {'spectrum': shape.sizes['spectrum'], **direct_beam.sizes}
    return DirectBeam(direct_beam.broadcast(sizes=sizes).copy())


@pytest.mark.parametrize(
    'uncertainties',
    [UncertaintyBroadcastMode.drop, UncertaintyBroadcastMode.upper_bound],
)
def test_pixel_dependent_direct_beam_is_supported(uncertainties):
    params = make_params()
    params[UncertaintyBroadcastMode] = uncertainties
    providers = as_dict(sans2d_providers())
    providers[DirectBeam] = pixel_dependent_direct_beam
    pipeline = sciline.Pipeline(providers.values(), params=params)
    result = pipeline.compute(BackgroundSubtractedIofQ)
    assert result.dims == ('Q',)


MANTID_BEAM_CENTER = sc.vector([0.09288, -0.08195, 0], unit='m')


def test_beam_center_from_center_of_mass_is_close_to_verified_result():
    params = make_params()
    providers = sans2d_providers()
    pipeline = sciline.Pipeline(providers, params=params)
    center = pipeline.compute(BeamCenter)
    # This is the result obtained from Mantid, using the full IofQ
    # calculation. The difference is about 3 mm in X or Y, probably due to a bias
    # introduced by the sample holder, which the center-of-mass approach cannot ignore.
    assert sc.allclose(center, MANTID_BEAM_CENTER, atol=sc.scalar(3e-3, unit='m'))


def test_beam_center_finder_without_direct_beam_reproduces_verified_result():
    params = make_params()
    params[sans.beam_center_finder.BeamCenterFinderQBins] = sc.linspace(
        'Q', 0.02, 0.3, 71, unit='1/angstrom'
    )
    del params[DirectBeamFilename]
    providers = sans2d_providers()
    providers.remove(sans.beam_center_finder.beam_center_from_center_of_mass)
    providers.append(sans.beam_center_finder.beam_center_from_iofq)
    pipeline = sciline.Pipeline(providers, params=params)
    pipeline[sans.beam_center_finder.BeamCenterFinderMinimizer] = None
    pipeline[sans.beam_center_finder.BeamCenterFinderTolerance] = None
    pipeline[DirectBeam] = None
    center = pipeline.compute(BeamCenter)
    assert sc.allclose(center, MANTID_BEAM_CENTER, atol=sc.scalar(2e-3, unit='m'))


def test_beam_center_can_get_closer_to_verified_result_with_low_counts_mask():
    def low_counts_mask(
        sample: RawData[SampleRun],
        low_counts_threshold: sans2d.LowCountThreshold,
    ) -> sans2d.SampleHolderMask:
        return sans2d.SampleHolderMask(sample.hist().data < low_counts_threshold)

    params = make_params()
    params[sans2d.LowCountThreshold] = sc.scalar(80.0, unit='counts')
    params[sans.beam_center_finder.BeamCenterFinderQBins] = sc.linspace(
        'Q', 0.02, 0.3, 71, unit='1/angstrom'
    )
    del params[DirectBeamFilename]
    providers = sans2d_providers()
    providers.remove(sans2d.sample_holder_mask)
    providers.remove(sans.beam_center_finder.beam_center_from_center_of_mass)
    providers.append(sans.beam_center_finder.beam_center_from_iofq)
    providers.append(low_counts_mask)
    pipeline = sciline.Pipeline(providers, params=params)
    pipeline[sans.beam_center_finder.BeamCenterFinderMinimizer] = None
    pipeline[sans.beam_center_finder.BeamCenterFinderTolerance] = None
    pipeline[DirectBeam] = None
    center = pipeline.compute(BeamCenter)
    assert sc.allclose(center, MANTID_BEAM_CENTER, atol=sc.scalar(5e-4, unit='m'))


def test_beam_center_finder_works_with_direct_beam():
    params = make_params()
    params[sans.beam_center_finder.BeamCenterFinderQBins] = sc.linspace(
        'Q', 0.02, 0.3, 71, unit='1/angstrom'
    )
    providers = sans2d_providers()
    providers.remove(sans.beam_center_finder.beam_center_from_center_of_mass)
    providers.append(sans.beam_center_finder.beam_center_from_iofq)
    pipeline = sciline.Pipeline(providers, params=params)
    pipeline[sans.beam_center_finder.BeamCenterFinderMinimizer] = None
    pipeline[sans.beam_center_finder.BeamCenterFinderTolerance] = None
    center_with_direct_beam = pipeline.compute(BeamCenter)
    assert sc.allclose(
        center_with_direct_beam, MANTID_BEAM_CENTER, atol=sc.scalar(2e-3, unit='m')
    )


def test_beam_center_finder_works_with_pixel_dependent_direct_beam():
    params = make_params()
    params[sans.beam_center_finder.BeamCenterFinderQBins] = sc.linspace(
        'Q', 0.02, 0.3, 71, unit='1/angstrom'
    )
    providers = sans2d_providers()
    providers.remove(sans.beam_center_finder.beam_center_from_center_of_mass)
    providers.append(sans.beam_center_finder.beam_center_from_iofq)
    pipeline = sciline.Pipeline(providers, params=params)
    pipeline[sans.beam_center_finder.BeamCenterFinderMinimizer] = None
    pipeline[sans.beam_center_finder.BeamCenterFinderTolerance] = None
    center_pixel_independent_direct_beam = pipeline.compute(BeamCenter)

    direct_beam = pipeline.compute(DirectBeam)
    pixel_dependent_direct_beam = direct_beam.broadcast(
        sizes={
            'spectrum': pipeline.compute(MaskedData[SampleRun]).sizes['spectrum'],
            'wavelength': direct_beam.sizes['wavelength'],
        }
    ).copy()

    providers = sans2d_providers()
    providers.remove(sans.beam_center_finder.beam_center_from_center_of_mass)
    providers.append(sans.beam_center_finder.beam_center_from_iofq)
    pipeline = sciline.Pipeline(providers, params=params)
    pipeline[sans.beam_center_finder.BeamCenterFinderMinimizer] = None
    pipeline[sans.beam_center_finder.BeamCenterFinderTolerance] = None
    pipeline[DirectBeam] = pixel_dependent_direct_beam

    center = pipeline.compute(BeamCenter)
    assert sc.identical(center, center_pixel_independent_direct_beam)


def test_workflow_runs_without_gravity_if_beam_center_is_provided():
    params = make_params()
    params[CorrectForGravity] = False
    pipeline = sciline.Pipeline(sans2d_providers(), params=params)
    da = pipeline.compute(RawData[SampleRun])
    del da.coords['gravity']
    pipeline[RawData[SampleRun]] = da
    pipeline[BeamCenter] = MANTID_BEAM_CENTER
    result = pipeline.compute(BackgroundSubtractedIofQ)
    assert result.dims == ('Q',)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from pathlib import Path

from ess.sans.data import Registry
from ess.sans.types import (
    BackgroundRun,
    Filename,
    PixelMaskFilename,
    SampleRun,
    TransmissionRun,
)

_registry = Registry(
    instrument='loki',
    files={
        # Files from LoKI@Larmor detector test experiment
        #
        # Background run 1 (no sample, sample holder/can only, no transmission monitor)
        '60248-2022-02-28_2215.nxs': 'md5:d9f17b95274a0fc6468df7e39df5bf03',
        # Sample run 1 (sample + sample holder/can, no transmission monitor in beam)
        '60250-2022-02-28_2215.nxs': 'md5:6a519ceaacbae702a6d08241e86799b1',
        # Sample run 2 (sample + sample holder/can, no transmission monitor in beam)
        '60339-2022-02-28_2215.nxs': 'md5:03c86f6389566326bb0cbbd80b8f8c4f',
        # Background transmission run (sample holder/can + transmission monitor)
        '60392-2022-02-28_2215.nxs': 'md5:9ecc1a9a2c05a880144afb299fc11042',
        # Background run 2 (no sample, sample holder/can only, no transmission monitor)
        '60393-2022-02-28_2215.nxs': 'md5:bf550d0ba29931f11b7450144f658652',
        # Sample transmission run (sample + sample holder/can + transmission monitor)
        '60394-2022-02-28_2215.nxs': 'md5:c40f38a62337d86957af925296c4c615',
        # Analytical model for the I(Q) of the Poly-Gauss sample
        'PolyGauss_I0-50_Rg-60.h5': 'md5:f5d60d9c2286cb197b8cd4dc82db3d7e',
        # XML file for the pixel mask
        'mask_new_July2022.xml': 'md5:421b6dc9db74126ffbc5d88164d017b0',
    },
    version='2',
)


def loki_tutorial_sample_run_60250() -> Filename[SampleRun]:
    """Sample run with sample and sample holder/can, no transmission monitor in beam."""
    return Filename[SampleRun](_registry.get_path('60250-2022-02-28_2215.nxs'))


def loki_tutorial_sample_run_60339() -> Filename[SampleRun]:
    """Sample run with sample and sample holder/can, no transmission monitor in beam."""
    return Filename[SampleRun](_registry.get_path('60339-2022-02-28_2215.nxs'))


def loki_tutorial_background_run_60248() -> Filename[BackgroundRun]:
    """Background run with sample holder/can only, no transmission monitor."""
    return Filename[BackgroundRun](_registry.get_path('60248-2022-02-28_2215.nxs'))


def loki_tutorial_background_run_60393() -> Filename[BackgroundRun]:
    """Background run with sample holder/can only, no transmission monitor."""
    return Filename[BackgroundRun](_registry.get_path('60393-2022-02-28_2215.nxs'))


def loki_tutorial_sample_transmission_run() -> Filename[TransmissionRun[SampleRun]]:
    """Sample transmission run (sample + sample holder/can + transmission monitor)."""
    return Filename[TransmissionRun[SampleRun]](
        _registry.get_path('60394-2022-02-28_2215.nxs')
    )


def loki_tutorial_run_60392() -> Filename[TransmissionRun[BackgroundRun]]:
    """Background transmission run (sample holder/can + transmission monitor), also
    used as empty beam run."""
    return Filename[TransmissionRun[BackgroundRun]](
        _registry.get_path('60392-2022-02-28_2215.nxs')
    )


def loki_tutorial_mask_filenames() -> list[PixelMaskFilename]:
    """List of pixel mask filenames for the LoKI@Larmor detector test experiment."""
    return [
        PixelMaskFilename(_registry.get_path('mask_new_July2022.xml')),
    ]


def loki_tutorial_poly_gauss_I0() -> Path:
    """Analytical model for the I(Q) of the Poly-Gauss sample."""
    return Path(_registry.get_path('PolyGauss_I0-50_Rg-60.h5'))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""
Default parameters, providers and utility functions for the loki workflow.
"""
from typing import Optional

import sciline
import scipp as sc
from ess.reduce import nexus

from ess.sans import providers as sans_providers

from ..sans.common import gravity_vector
from ..sans.types import (
    ConfiguredReducibleData,
    ConfiguredReducibleMonitor,
    CorrectForGravity,
    DetectorPixelShape,
    DimsToKeep,
    Incident,
    LabFrameTransform,
    LoadedNeXusDetector,
    LoadedNeXusMonitor,
    MonitorType,
    NeXusDetectorName,
    NeXusMonitorName,
    NonBackgroundWavelengthRange,
    PixelShapePath,
    RawData,
    RawMonitor,
    RawSample,
    RawSource,
    RunType,
    SamplePosition,
    ScatteringRunType,
    SourcePosition,
    TofData,
    TofMonitor,
    TransformationPath,
    Transmission,
    WavelengthBands,
    WavelengthMask,
)
from .io import dummy_load_sample


def default_parameters() -> dict:
    return {
        CorrectForGravity: False,
        DimsToKeep: tuple(),
        NeXusMonitorName[Incident]: 'monitor_1',
        NeXusMonitorName[Transmission]: 'monitor_2',
        TransformationPath: 'transform',
        PixelShapePath: 'pixel_shape',
        NonBackgroundWavelengthRange: None,
        WavelengthMask: None,
        WavelengthBands: None,
    }


def LokiAtLarmorWorkflow() -> sciline.Pipeline:
    """
    Workflow with default parameters for Loki test at Larmor.

    This version of the Loki workflow:

    - Uses ISIS XML files to define masks.
    - Sets a dummy sample position [0,0,0] since files do not contain this information.

    Returns
    -------
    :
        Loki workflow as a sciline.Pipeline
    """
    from ess.isissans.io import read_xml_detector_masking

    from . import providers as loki_providers

    params = default_parameters()
    loki_providers = sans_providers + loki_providers
    workflow = sciline.Pipeline(providers=loki_providers, params=params)
    workflow.insert(read_xml_detector_masking)
    # No sample information in the Loki@Larmor files, so we use a dummy sample provider
    workflow.insert(dummy_load_sample)
    return workflow


DETECTOR_BANK_RESHAPING = {
    'larmor_detector': lambda x: x.fold(
        dim='detector_number', sizes=dict(layer=4, tube=32, straw=7, pixel=512)
    )
}


def get_source_position(
    raw_source: RawSource[RunType],
) -> SourcePosition[RunType]:
    return SourcePosition[RunType](raw_source['position'])


def get_sample_position(
    raw_sample: RawSample[RunType],
) -> SamplePosition[RunType]:
    return SamplePosition[RunType](raw_sample['position'])


def get_detector_data(
    detector: LoadedNeXusDetector[ScatteringRunType],
    detector_name: NeXusDetectorName,
) -> RawData[ScatteringRunType]:
    da = nexus.extract_detector_data(detector)
    if detector_name in DETECTOR_BANK_RESHAPING:
        da = DETECTOR_BANK_RESHAPING[detector_name](da)
    return RawData[ScatteringRunType](da)


def get_monitor_data(
    monitor: LoadedNeXusMonitor[RunType, MonitorType],
) -> RawMonitor[RunType, MonitorType]:
    out = nexus.extract_monitor_data(monitor).copy(deep=False)
    out.coords['position'] = monitor['position']
    return RawMonitor[RunType, MonitorType](out)


def _add_variances_and_coordinates(
    da: sc.DataArray,
    source_position: sc.Variable,
    sample_position: Optional[sc.Variable] = None,
) -> sc.DataArray:
    out = da.copy(deep=False)
    if out.bins is not None:
        content = out.bins.constituents['data']
        if content.variances is None:
            content.variances = content.values
    # Sample position is not needed in the case of a monitor.
    if sample_position is not None:
        out.coords['sample_position'] = sample_position
    out.coords['source_position'] = source_position
    out.coords['gravity'] = gravity_vector()
    return out


def patch_detector_data(
    detector_data: RawData[ScatteringRunType],
    source_position: SourcePosition[ScatteringRunType],
    sample_position: SamplePosition[ScatteringRunType],
) -> ConfiguredReducibleData[ScatteringRunType]:
    return ConfiguredReducibleData[ScatteringRunType](
        _add_variances_and_coordinates(
            da=detector_data,
            source_position=source_position,
            sample_position=sample_position,
        )
    )


def patch_monitor_data(
    monitor_data: RawMonitor[RunType, MonitorType],
    source_position: SourcePosition[RunType],
) -> ConfiguredReducibleMonitor[RunType, MonitorType]:
    return ConfiguredReducibleMonitor[RunType, MonitorType](
        _add_variances_and_coordinates(da=monitor_data, source_position=source_position)
    )


def _convert_to_tof(da: sc.DataArray) -> sc.DataArray:
    da.bins.coords['tof'] = da.bins.coords.pop('event_time_offset')
    if 'event_time_zero' in da.dims:
        da = da.bins.concat('event_time_zero')
    return da


def data_to_tof(
    da: ConfiguredReducibleData[ScatteringRunType],
) -> TofData[ScatteringRunType]:
    return TofData[ScatteringRunType](_convert_to_tof(da))


def monitor_to_tof(
    da: ConfiguredReducibleMonitor[RunType, MonitorType],
) -> TofMonitor[RunType, MonitorType]:
    return TofMonitor[RunType, MonitorType](_convert_to_tof(da))


def detector_pixel_shape(
    detector: LoadedNeXusDetector[ScatteringRunType],
    pixel_shape_path: PixelShapePath,
) -> DetectorPixelShape[ScatteringRunType]:
    return DetectorPixelShape[ScatteringRunType](detector[pixel_shape_path])


def detector_lab_frame_transform(
    detector: LoadedNeXusDetector[ScatteringRunType],
    transform_path: TransformationPath,
) -> LabFrameTransform[ScatteringRunType]:
    return LabFrameTransform[ScatteringRunType](detector[transform_path])


providers = (
    detector_pixel_shape,
    detector_lab_frame_transform,
    get_detector_data,
    get_monitor_data,
    get_sample_position,
    get_source_position,
    patch_detector_data,
    patch_monitor_data,
    data_to_tof,
    monitor_to_tof,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import os
import sys
from pathlib import Path

import pytest
import sciline
import scipp as sc
from scipp.testing import assert_identical

from ess import loki, sans
from ess.sans.conversions import ElasticCoordTransformGraph
from ess.sans.types import (
    BackgroundSubtractedIofQ,
    BackgroundSubtractedIofQxy,
    BeamCenter,
    CalibratedMaskedData,
    CleanSummedQ,
    CleanWavelengthMasked,
    CorrectForGravity,
    Denominator,
    DimsToKeep,
    IofQ,
    IofQxy,
    NeXusDetectorName,
    Numerator,
    QBins,
    QxBins,
    QyBins,
    ReturnEvents,
    SampleRun,
    UncertaintyBroadcastMode,
    WavelengthBands,
    WavelengthBins,
)

sys.path.insert(0, str(Path(__file__).resolve().parent))
from common import (  # noqa: E402
    loki_providers,
    loki_providers_no_beam_center_finder,
    make_params,
)


def test_can_create_pipeline():
    pipeline = sciline.Pipeline(loki_providers(), params=make_params())
    pipeline.get(BackgroundSubtractedIofQ)


def test_can_create_pipeline_with_pixel_masks():
    pipeline = sciline.Pipeline(loki_providers(), params=make_params(no_masks=False))
    pipeline = sans.with_pixel_mask_filenames(
        pipeline, loki.data.loki_tutorial_mask_filenames()
    )
    pipeline.get(BackgroundSubtractedIofQ)


@pytest.mark.parametrize(
    'uncertainties',
    [UncertaintyBroadcastMode.drop, UncertaintyBroadcastMode.upper_bound],
)
@pytest.mark.parametrize('qxy', [False, True])
def test_pipeline_can_compute_IofQ(uncertainties, qxy: bool):
    params = make_params(no_masks=False)
    params[UncertaintyBroadcastMode] = uncertainties
    pipeline = sciline.Pipeline(loki_providers(), params=params)
    pipeline = sans.with_pixel_mask_filenames(
        pipeline, loki.data.loki_tutorial_mask_filenames()
    )
    if qxy:
        result = pipeline.compute(BackgroundSubtractedIofQxy)
        assert result.dims == ('Qy', 'Qx')
        assert sc.identical(result.coords['Qx'], params[QxBins])
        assert sc.identical(result.coords['Qy'], params[QyBins])
        assert result.sizes['Qx'] == 90
        assert result.sizes['Qy'] == 77
    else:
        result = pipeline.compute(BackgroundSubtractedIofQ)
        assert result.dims == ('Q',)
        assert sc.identical(result.coords['Q'], params[QBins])
        assert result.sizes['Q'] == 100
    if uncertainties == UncertaintyBroadcastMode.drop:
        test_dir = os.path.dirname(os.path.abspath(__file__))
        name = Path(f'reference_IofQ{"xy" if qxy else ""}_{uncertainties}.hdf5')
        reference = sc.io.load_hdf5(test_dir / name)
        assert_identical(result, reference)


@pytest.mark.parametrize(
    'uncertainties',
    [UncertaintyBroadcastMode.drop, UncertaintyBroadcastMode.upper_bound],
)
@pytest.mark.parametrize(
    'target',
    [
        IofQ[SampleRun],
        IofQxy[SampleRun],
        BackgroundSubtractedIofQ,
        BackgroundSubtractedIofQxy,
    ],
)
def test_pipeline_can_compute_IofQ_in_event_mode(uncertainties, target):
    params = make_params()
    params[UncertaintyBroadcastMode] = uncertainties
    pipeline = sciline.Pipeline(loki_providers(), params=params)
    reference = pipeline.compute(target)
    pipeline[ReturnEvents] = True
    result = pipeline.compute(target)
    qxy = target in (IofQxy[SampleRun], BackgroundSubtractedIofQxy)
    assert result.bins is not None
    assert result.dims == ('Qy', 'Qx') if qxy else ('Q',)
    assert sc.allclose(
        sc.values(reference.data),
        sc.values(result.hist().data),
        # Could be 1e-11, but currently the workflow defaults to float32 data, as
        # returned by ScippNexus.
        rtol=sc.scalar(1e-7),
        atol=sc.scalar(1e-7),
    )
    if uncertainties == UncertaintyBroadcastMode.drop:
        # Could both be 1e-14 if using float64
        tol = sc.scalar(1e-7 if qxy else 1e-10)
    else:
        # Could be 1e-8 if using float64
        tol = sc.scalar(1e-7 if qxy else 1e-9)
    assert sc.allclose(
        sc.variances(reference).data,
        sc.variances(result.hist()).data,
        rtol=tol,
        atol=tol,
    )


@pytest.mark.parametrize('qxy', [False, True])
def test_pipeline_can_compute_IofQ_in_wavelength_bands(qxy: bool):
    params = make_params()
    params[WavelengthBands] = sc.linspace(
        'wavelength',
        params[WavelengthBins].min(),
        params[WavelengthBins].max(),
        11,
    )
    pipeline = sciline.Pipeline(loki_providers(), params=params)
    result = pipeline.compute(
        BackgroundSubtractedIofQxy if qxy else BackgroundSubtractedIofQ
    )
    assert result.dims == ('band', 'Qy', 'Qx') if qxy else ('band', 'Q')
    assert result.sizes['band'] == 10


@pytest.mark.parametrize('qxy', [False, True])
def test_pipeline_can_compute_IofQ_in_overlapping_wavelength_bands(qxy: bool):
    params = make_params()
    # Bands have double the width
    edges = sc.linspace(
        'band', params[WavelengthBins].min(), params[WavelengthBins].max(), 12
    )
    params[WavelengthBands] = sc.concat(
        [edges[:-2], edges[2::]], dim='wavelength'
    ).transpose()
    pipeline = sciline.Pipeline(loki_providers(), params=params)
    result = pipeline.compute(
        BackgroundSubtractedIofQxy if qxy else BackgroundSubtractedIofQ
    )
    assert result.dims == ('band', 'Qy', 'Qx') if qxy else ('band', 'Q')
    assert result.sizes['band'] == 10


@pytest.mark.parametrize('qxy', [False, True])
def test_pipeline_can_compute_IofQ_in_layers(qxy: bool):
    params = make_params()
    params[DimsToKeep] = ['layer']
    pipeline = sciline.Pipeline(loki_providers(), params=params)
    result = pipeline.compute(
        BackgroundSubtractedIofQxy if qxy else BackgroundSubtractedIofQ
    )
    assert result.dims == ('layer', 'Qy', 'Qx') if qxy else ('layer', 'Q')
    assert result.sizes['layer'] == 4


def _compute_beam_center():
    pipeline = sciline.Pipeline(loki_providers(), params=make_params())
    center = pipeline.compute(BeamCenter)
    return center


def test_pipeline_can_compute_IofQ_merging_events_from_multiple_runs():
    params = make_params()

    sample_runs = [
        loki.data.loki_tutorial_sample_run_60250(),
        loki.data.loki_tutorial_sample_run_60339(),
    ]
    background_runs = [
        loki.data.loki_tutorial_background_run_60248(),
        loki.data.loki_tutorial_background_run_60393(),
    ]
    pipeline = sciline.Pipeline(loki_providers_no_beam_center_finder(), params=params)
    pipeline[BeamCenter] = _compute_beam_center()

    pipeline = sans.with_sample_runs(pipeline, runs=sample_runs)
    pipeline = sans.with_background_runs(pipeline, runs=background_runs)

    result = pipeline.compute(BackgroundSubtractedIofQ)
    assert result.dims == ('Q',)


def test_pipeline_can_compute_IofQ_merging_events_from_banks():
    params = make_params()
    del params[NeXusDetectorName]

    pipeline = sciline.Pipeline(loki_providers_no_beam_center_finder(), params=params)
    pipeline[BeamCenter] = _compute_beam_center()
    pipeline = sans.with_banks(pipeline, banks=['larmor_detector'])

    result = pipeline.compute(BackgroundSubtractedIofQ)
    assert result.dims == ('Q',)


def test_pipeline_can_compute_IofQ_merging_events_from_multiple_runs_and_banks():
    params = make_params()
    sample_runs = [
        loki.data.loki_tutorial_sample_run_60250(),
        loki.data.loki_tutorial_sample_run_60339(),
    ]
    background_runs = [
        loki.data.loki_tutorial_background_run_60248(),
        loki.data.loki_tutorial_background_run_60393(),
    ]
    pipeline = sciline.Pipeline(loki_providers_no_beam_center_finder(), params=params)
    pipeline[BeamCenter] = _compute_beam_center()

    pipeline = sans.with_sample_runs(pipeline, runs=sample_runs)
    pipeline = sans.with_background_runs(pipeline, runs=background_runs)
    key = BackgroundSubtractedIofQ
    reference = pipeline.compute(key)

    pipeline = sans.with_banks(pipeline, banks=['larmor_detector', 'larmor_detector'])
    result = pipeline.compute(key)

    # Note that the variances are not the same for the bank-merged data since we use
    # the same detector twice.
    assert_identical(sc.values(result), sc.values(reference))


def test_pipeline_IofQ_merging_events_yields_consistent_results():
    N = 3
    params = make_params()
    center = _compute_beam_center()
    pipeline_single = sciline.Pipeline(
        loki_providers_no_beam_center_finder(), params=params
    )
    pipeline_single[BeamCenter] = center

    sample_runs = [loki.data.loki_tutorial_sample_run_60339()] * N
    background_runs = [loki.data.loki_tutorial_background_run_60393()] * N
    pipeline_triple = sans.with_sample_runs(pipeline_single, runs=sample_runs)
    pipeline_triple = sans.with_background_runs(pipeline_triple, runs=background_runs)

    iofq1 = pipeline_single.compute(BackgroundSubtractedIofQ)
    iofq3 = pipeline_triple.compute(BackgroundSubtractedIofQ)
    assert sc.allclose(sc.values(iofq1.data), sc.values(iofq3.data))
    assert sc.identical(iofq1.coords['Q'], iofq3.coords['Q'])
    assert all(sc.variances(iofq1.data) > sc.variances(iofq3.data))
    assert sc.allclose(
        sc.values(
            pipeline_single.compute(CleanSummedQ[SampleRun, Numerator]).hist().data
        )
        * N,
        sc.values(
            pipeline_triple.compute(CleanSummedQ[SampleRun, Numerator]).hist().data
        ),
    )
    assert sc.allclose(
        sc.values(pipeline_single.compute(CleanSummedQ[SampleRun, Denominator]).data)
        * N,
        sc.values(pipeline_triple.compute(CleanSummedQ[SampleRun, Denominator]).data),
    )


def test_beam_center_from_center_of_mass_is_close_to_verified_result():
    params = make_params(no_masks=False)
    pipeline = sciline.Pipeline(loki_providers(), params=params)
    pipeline = sans.with_pixel_mask_filenames(
        pipeline, loki.data.loki_tutorial_mask_filenames()
    )
    center = pipeline.compute(BeamCenter)
    reference = sc.vector([-0.0291487, -0.0181614, 0], unit='m')
    assert sc.allclose(center, reference)


def test_phi_with_gravity():
    params = make_params()
    pipeline = sciline.Pipeline(loki_providers(), params=params)
    pipeline[CorrectForGravity] = False
    data_no_grav = pipeline.compute(
        CleanWavelengthMasked[SampleRun, Numerator]
    ).flatten(to='pixel')
    graph_no_grav = pipeline.compute(ElasticCoordTransformGraph)
    pipeline[CorrectForGravity] = True
    data_with_grav = (
        pipeline.compute(CleanWavelengthMasked[SampleRun, Numerator])
        .flatten(to='pixel')
        .hist(wavelength=sc.linspace('wavelength', 1.0, 12.0, 101, unit='angstrom'))
    )
    graph_with_grav = pipeline.compute(ElasticCoordTransformGraph)

    no_grav = data_no_grav.transform_coords(('two_theta', 'phi'), graph_no_grav)
    phi_no_grav = no_grav.coords['phi']
    with_grav = data_with_grav.transform_coords(('two_theta', 'phi'), graph_with_grav)
    phi_with_grav = with_grav.coords['phi'].mean('wavelength')

    assert not sc.identical(phi_no_grav, phi_with_grav)

    # Exclude pixels near y=0, since phi with gravity could drop below y=0 and give a
    # difference of almost 2*pi.
    y = sc.abs(
        pipeline.compute(CalibratedMaskedData[SampleRun])
        .coords['position']
        .fields.y.flatten(to='pixel')
    )
    not_near_origin = y > sc.scalar(0.05, unit='m')
    phi_no_grav = phi_no_grav[not_near_origin]
    phi_with_grav = phi_with_grav[not_near_origin]
    assert sc.all(
        sc.isclose(phi_no_grav, phi_with_grav, atol=sc.scalar(5.0e-3, unit='rad'))
    )

    # Phi is in [-pi, pi], measured from the X axis.
    pos_x = sc.abs(phi_no_grav) < sc.scalar(90.0, unit='deg').to(unit='rad')
    # Phi is larger with gravity, since it gives the position where it would have
    # been detected without gravity. That is, with gravity all points are pulled
    # "up" in the XY plane, so the angle is larger for positive X and smaller for
    # negative X.
    assert sc.all(phi_no_grav[pos_x] < phi_with_grav[pos_x])
    assert sc.all(phi_no_grav[~pos_x] > phi_with_grav[~pos_x])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
"""
File loading functions for ISIS data, NOT using Mantid.
"""
from typing import NewType

import scipp as sc

from ess.sans.types import MaskedDetectorIDs, PixelMaskFilename

CalibrationFilename = NewType('CalibrationFilename', str)


def read_xml_detector_masking(filename: PixelMaskFilename) -> MaskedDetectorIDs:
    """Read a pixel mask from an XML file.

    The format is as follows, where the detids are inclusive ranges of detector IDs:

    .. code-block:: xml

        <?xml version="1.0"?>
        <detector-masking>
            <group>
                <detids>1400203-1400218,1401199,1402190-1402223</detids>
            </group>
        </detector-masking>

    Parameters
    ----------
    filename:
        Path to the XML file.
    """
    import xml.etree.ElementTree as ET  # nosec

    tree = ET.parse(filename)  # nosec
    root = tree.getroot()

    masked_detids = []
    for group in root.findall('group'):
        for detids in group.findall('detids'):
            for detid in detids.text.split(','):
                detid = detid.strip()
                if '-' in detid:
                    start, stop = detid.split('-')
                    masked_detids += list(range(int(start), int(stop) + 1))
                else:
                    masked_detids.append(int(detid))

    return MaskedDetectorIDs(
        sc.array(dims=['detector_id'], values=masked_detids, unit=None, dtype='int32')
    )


providers = (read_xml_detector_masking,)



# This file is used by beamlime to create a workflow for the Loki instrument.
# The callable `live_workflow` is registered as the entry point for the workflow.
import sciline
import scipp as sc
import scippnexus as snx
from scippneutron.io.nexus.load_nexus import JSONGroup

from ess.loki.general import (
    get_monitor_data,
    get_source_position,
    monitor_to_tof,
    patch_monitor_data,
)
from ess.loki.io import load_nexus_monitor, load_nexus_source
from ess.sans.conversions import monitor_to_wavelength, sans_monitor
from ess.sans.types import (
    Filename,
    Incident,
    MonitorType,
    NeXusMonitorName,
    RunType,
    SampleRun,
    Transmission,
    WavelengthBins,
    WavelengthMonitor,
)


class MonitorHistogram(
    sciline.ScopeTwoParams[RunType, MonitorType, sc.DataArray], sc.DataArray
):
    ...


def _hist_monitor_wavelength(
    wavelength_bin: WavelengthBins, monitor: WavelengthMonitor[RunType, MonitorType]
) -> MonitorHistogram[RunType, MonitorType]:
    return monitor.hist(wavelength=wavelength_bin)


def loki_iofq_workflow(group: JSONGroup) -> dict[str, sc.DataArray]:
    """Example live workflow function for Loki.

    This function is used to process data with beamlime workflows.
    """
    # We do not consume the incoming data right now.
    _ = group
    return {
        'IofQ': sc.DataArray(
            data=sc.zeros(dims=['Q'], shape=[100]),
            coords={
                'Q': sc.linspace(
                    dim='Q', start=0.01, stop=0.3, num=101, unit='1/angstrom'
                )
            },
        )
    }


class LoKiMonitorWorkflow:
    """LoKi Monitor wavelength histogram workflow for live data reduction."""

    def __init__(self) -> None:
        self.pipeline = self._build_pipeline()

    def _build_pipeline(self) -> sciline.Pipeline:
        """Build a workflow pipeline for live data reduction.

        Returns
        -------
        :
            A pipeline for live data reduction.
            The initial pipeline will be missing the input data.
            It should be set before calling the pipeline.

        """
        # Wavelength binning parameters
        wavelength_min = sc.scalar(1.0, unit="angstrom")
        wavelength_max = sc.scalar(13.0, unit="angstrom")
        n_wavelength_bins = 50

        providers = (
            load_nexus_monitor,
            load_nexus_source,
            get_source_position,
            get_monitor_data,
            patch_monitor_data,
            monitor_to_tof,
            sans_monitor,
            monitor_to_wavelength,
            _hist_monitor_wavelength,
        )

        params = {
            NeXusMonitorName[Incident]: "monitor_1",
            NeXusMonitorName[Transmission]: "monitor_2",
            WavelengthBins: sc.linspace(
                "wavelength", wavelength_min, wavelength_max, n_wavelength_bins + 1
            ),
        }
        workflow = sciline.Pipeline(providers, params=params)
        return workflow

    def __call__(self, group: JSONGroup) -> dict[str, sc.DataArray]:
        """

        Returns
        -------
        :
            Plottable Outputs:

            - MonitorHistogram[SampleRun, Incident]
            - MonitorHistogram[SampleRun, Transmission]

        """
        # ``JsonGroup`` is turned into the ``NexusGroup`` here, not in the ``beamlime``
        # so that the workflow can control the definition of the group.
        self.pipeline[Filename[SampleRun]] = snx.Group(
            group, definitions=snx.base_definitions()
        )
        results = self.pipeline.compute(
            (
                MonitorHistogram[SampleRun, Incident],
                MonitorHistogram[SampleRun, Transmission],
            )
        )

        return {str(tp): result for tp, result in results.items()}


loki_monitor_workflow = LoKiMonitorWorkflow()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""
This modules defines the domain types uses in esssans.

The domain types are used to define parameters and to request results from a Sciline
pipeline."""
from collections.abc import Sequence
from enum import Enum
from typing import NewType, TypeVar

import sciline
import scipp as sc

# 1  TypeVars used to parametrize the generic parts of the workflow

# 1.1  Run types
BackgroundRun = NewType('BackgroundRun', int)
"""Background run: the run with only the solvent which the sample is placed in."""
EmptyBeamRun = NewType('EmptyBeamRun', int)
"""Run (sometimes called 'direct run') where the sample holder was empty.
It is used for reading the data from the transmission monitor."""
SampleRun = NewType('SampleRun', int)
"""Sample run: the run with the sample placed in the solvent inside the sample holder.
"""

ScatteringRunType = TypeVar(
    'ScatteringRunType',
    SampleRun,
    BackgroundRun,
)


class TransmissionRun(sciline.Scope[ScatteringRunType, int], int):
    """Mapping between ScatteringRunType and transmission run.
    In the case where no transmission run is provided, the transmission run should be
    the same as the measurement (sample or background) run."""


RunType = TypeVar(
    'RunType',
    BackgroundRun,
    EmptyBeamRun,
    SampleRun,
    # Note that mypy does not seem to like this nesting, may need to find a workaround
    TransmissionRun[SampleRun],
    TransmissionRun[BackgroundRun],
)
"""TypeVar used for specifying BackgroundRun, EmptyBeamRun or SampleRun"""

# 1.2  Monitor types
Incident = NewType('Incident', int)
"""Incident monitor"""
Transmission = NewType('Transmission', int)
"""Transmission monitor"""
MonitorType = TypeVar('MonitorType', Incident, Transmission)
"""TypeVar used for specifying Incident or Transmission monitor type"""

# 1.3  Numerator and denominator of IofQ
Numerator = NewType('Numerator', sc.DataArray)
"""Numerator of IofQ"""
Denominator = NewType('Denominator', sc.DataArray)
"""Denominator of IofQ"""
IofQPart = TypeVar('IofQPart', Numerator, Denominator)
"""TypeVar used for specifying Numerator or Denominator of IofQ"""

# 1.4  Entry paths in NeXus files
NeXusDetectorName = NewType('NeXusDetectorName', str)
"""Name of detector entry in NeXus file"""

PixelShapePath = NewType('PixelShapePath', str)
"""
Name of the entry where the pixel shape is stored in the NeXus file
"""

TransformationPath = NewType('TransformationPath', str)
"""
Name of the entry where the transformation computed from the
transformation chain is stored, for the detectors and the monitors
"""

# 2  Workflow parameters

UncertaintyBroadcastMode = Enum(
    'UncertaintyBroadcastMode', ['drop', 'upper_bound', 'fail']
)
"""
Mode for broadcasting uncertainties.

See https://doi.org/10.3233/JNR-220049 for context.
"""

ReturnEvents = NewType('ReturnEvents', bool)
"""Whether to return events in the output I(Q)"""

WavelengthBins = NewType('WavelengthBins', sc.Variable)
"""Wavelength binning"""

WavelengthBands = NewType('WavelengthBands', sc.Variable | None)
"""Wavelength bands. Typically a single band, set to first and last value of
WavelengthBins.

The wavelength bands can however be used to compute the scattering cross section inside
multiple wavelength bands. In this case, the wavelength bands must be either one- or
two-dimensional.

In the case of a one-dimensional array, the values represent a set of non-overlapping
bins in the wavelength dimension.

In the case of a two-dimensional array, the values represent a set of (possibly
overlapping or non-contiguous) wavelength bands, with the first dimension being the
band index and the second dimension being the wavelength. For each band, there must be
two wavelength values defining the start and end wavelength of the band.
"""

ProcessedWavelengthBands = NewType('ProcessedWavelengthBands', sc.Variable)
"""Processed wavelength bands, as a two-dimensional variable, with the first dimension
being the band index and the second dimension being the wavelength. For each band, there
must be two wavelength values defining the start and end wavelength of the band."""


QBins = NewType('QBins', sc.Variable)
"""Q binning used when computing IofQ"""

QxBins = NewType('QxBins', sc.Variable)
"""Qx binning used when computing IofQxy"""

QyBins = NewType('QyBins', sc.Variable)
"""Qy binning used when computing IofQxy"""

NonBackgroundWavelengthRange = NewType(
    'NonBackgroundWavelengthRange', sc.Variable | None
)
"""Range of wavelengths that are not considered background in the monitor"""

DirectBeamFilename = NewType('DirectBeamFilename', str)
"""Filename of direct beam correction"""

BeamCenter = NewType('BeamCenter', sc.Variable)
"""Beam center, may be set directly or computed using beam-center finder"""

WavelengthMask = NewType('WavelengthMask', sc.DataArray | None)
"""Optional wavelength mask"""

CorrectForGravity = NewType('CorrectForGravity', bool)
"""Whether to correct for gravity when computing wavelength and Q."""

DimsToKeep = NewType('DimsToKeep', Sequence[str])
"""Dimensions that should not be reduced and thus still be present in the final
I(Q) result (this is typically the layer dimension)."""

OutFilename = NewType('OutFilename', str)
"""Filename of the output"""


class NeXusMonitorName(sciline.Scope[MonitorType, str], str):
    """Name of Incident|Transmission monitor in NeXus file"""


PixelMaskFilename = NewType('PixelMaskFilename', str)

FilenameType = TypeVar('FilenameType', bound=str)


class Filename(sciline.Scope[RunType, str], str):
    """Filename of a run"""


DetectorIDs = NewType('DetectorIDs', sc.Variable)
"""1-D variable listing all detector IDs."""

MaskedDetectorIDs = NewType('MaskedDetectorIDs', sc.Variable)
"""1-D variable listing all masked detector IDs."""


DetectorMasks = NewType('DetectorMasks', dict[str, sc.Variable])
"""Masks for detector pixels"""


# 3  Workflow (intermediate) results


class RawSource(sciline.Scope[RunType, sc.DataGroup], sc.DataGroup):
    """Raw source from NeXus file"""


class RawSample(sciline.Scope[RunType, sc.DataGroup], sc.DataGroup):
    """Raw sample from NeXus file"""


class SamplePosition(sciline.Scope[RunType, sc.Variable], sc.Variable):
    """Sample position"""


class SourcePosition(sciline.Scope[RunType, sc.Variable], sc.Variable):
    """Source position"""


DirectBeam = NewType('DirectBeam', sc.DataArray | None)
"""Direct beam"""


class TransmissionFraction(
    sciline.Scope[ScatteringRunType, sc.DataArray], sc.DataArray
):
    """Transmission fraction"""


CleanDirectBeam = NewType('CleanDirectBeam', sc.DataArray | None)
"""Direct beam after resampling to required wavelength bins"""


class DetectorPixelShape(sciline.Scope[ScatteringRunType, sc.DataGroup], sc.DataGroup):
    """Geometry of the detector from description in nexus file."""


class LabFrameTransform(sciline.Scope[ScatteringRunType, sc.Variable], sc.Variable):
    """Coordinate transformation from detector local coordinates
    to the sample frame of reference."""


class SolidAngle(sciline.Scope[ScatteringRunType, sc.DataArray], sc.DataArray):
    """Solid angle of detector pixels seen from sample position"""


class LoadedNeXusDetector(sciline.Scope[RunType, sc.DataGroup], sc.DataGroup):
    """Detector data, loaded from a NeXus file, containing not only neutron events
    but also pixel shape information, transformations, ..."""


class LoadedNeXusMonitor(
    sciline.ScopeTwoParams[RunType, MonitorType, sc.DataGroup], sc.DataGroup
):
    """Monitor data loaded from a NeXus file, containing not only neutron events
    but also transformations, ..."""


class RawData(sciline.Scope[ScatteringRunType, sc.DataArray], sc.DataArray):
    """Raw detector data"""


class ConfiguredReducibleData(
    sciline.Scope[ScatteringRunType, sc.DataArray], sc.DataArray
):
    """Raw event data where variances and necessary coordinates
    (e.g. sample and source position) have been added, and where optionally some
    user configuration was applied to some of the coordinates."""


class TofData(sciline.Scope[ScatteringRunType, sc.DataArray], sc.DataArray):
    """Data with a time-of-flight coordinate"""


class TofMonitor(
    sciline.ScopeTwoParams[RunType, MonitorType, sc.DataGroup], sc.DataGroup
):
    """Monitor data with a time-of-flight coordinate"""


PixelMask = NewType('PixelMask', sc.Variable)


class MaskedData(sciline.Scope[ScatteringRunType, sc.DataArray], sc.DataArray):
    """Raw data with pixel-specific masks applied"""


class CalibratedMaskedData(
    sciline.Scope[ScatteringRunType, sc.DataArray], sc.DataArray
):
    """Raw data with pixel-specific masks applied and calibrated pixel positions"""


class NormWavelengthTerm(sciline.Scope[ScatteringRunType, sc.DataArray], sc.DataArray):
    """Normalization term (numerator) for IofQ before scaling with solid-angle."""


class CleanWavelength(
    sciline.ScopeTwoParams[ScatteringRunType, IofQPart, sc.DataArray], sc.DataArray
):
    """
    Prerequisite for IofQ numerator or denominator.

    This can either be the sample or background counts, converted to wavelength,
    or the respective normalization terms computed from the respective solid angle,
    direct beam, and monitors.
    """


class CleanWavelengthMasked(
    sciline.ScopeTwoParams[ScatteringRunType, IofQPart, sc.DataArray], sc.DataArray
):
    """Result of applying wavelength masking to :py:class:`CleanWavelength`"""


class CleanQ(
    sciline.ScopeTwoParams[ScatteringRunType, IofQPart, sc.DataArray], sc.DataArray
):
    """Result of converting :py:class:`CleanWavelengthMasked` to Q"""


class CleanQxy(
    sciline.ScopeTwoParams[ScatteringRunType, IofQPart, sc.DataArray], sc.DataArray
):
    """Result of converting :py:class:`CleanWavelengthMasked` to Qx and Qy"""


class CleanSummedQ(
    sciline.ScopeTwoParams[ScatteringRunType, IofQPart, sc.DataArray], sc.DataArray
):
    """Result of histogramming/binning :py:class:`CleanQ` over all pixels into Q bins"""


class CleanSummedQxy(
    sciline.ScopeTwoParams[ScatteringRunType, IofQPart, sc.DataArray], sc.DataArray
):
    """Result of histogramming/binning :py:class:`CleanQxy` over all pixels into Qx and
    Qy bins"""


class IofQ(sciline.Scope[ScatteringRunType, sc.DataArray], sc.DataArray):
    """I(Q)"""


class IofQxy(sciline.Scope[ScatteringRunType, sc.DataArray], sc.DataArray):
    """I(Qx, Qy)"""


BackgroundSubtractedIofQ = NewType('BackgroundSubtractedIofQ', sc.DataArray)
"""I(Q) with background (given by I(Q) of the background run) subtracted"""

BackgroundSubtractedIofQxy = NewType('BackgroundSubtractedIofQxy', sc.DataArray)
"""I(Qx, Qy) with background (given by I(Qx, Qy) of the background run) subtracted"""


class RawMonitor(
    sciline.ScopeTwoParams[RunType, MonitorType, sc.DataArray], sc.DataArray
):
    """Raw monitor data"""


class ConfiguredReducibleMonitor(
    sciline.ScopeTwoParams[RunType, MonitorType, sc.DataArray], sc.DataArray
):
    """Raw monitor data where variances and necessary coordinates
    (e.g. source position) have been added, and where optionally some
    user configuration was applied to some of the coordinates."""


class WavelengthMonitor(
    sciline.ScopeTwoParams[RunType, MonitorType, sc.DataArray], sc.DataArray
):
    """Monitor data converted to wavelength"""


class CleanMonitor(
    sciline.ScopeTwoParams[RunType, MonitorType, sc.DataArray], sc.DataArray
):
    """Monitor data cleaned of background counts"""


# 4  Metadata

RunTitle = NewType('RunTitle', str)
"""Title of the run."""

RunNumber = NewType('RunNumber', int)
"""Run number."""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from typing import Hashable, Iterable

import pandas as pd
import sciline
import scipp as sc

from .types import (
    BackgroundRun,
    CleanSummedQ,
    Denominator,
    DetectorMasks,
    Filename,
    NeXusDetectorName,
    Numerator,
    PixelMaskFilename,
    SampleRun,
)


def _merge(*dicts: dict) -> dict:
    return {key: value for d in dicts for key, value in d.items()}


def merge_contributions(*data: sc.DataArray) -> sc.DataArray:
    if len(data) == 1:
        return data[0]
    reducer = sc.reduce(data)
    return reducer.bins.concat() if data[0].bins is not None else reducer.sum()


def with_pixel_mask_filenames(
    workflow: sciline.Pipeline, masks: Iterable[str]
) -> sciline.Pipeline:
    """
    Return modified workflow with pixel mask filenames set.

    Parameters
    ----------
    workflow:
        Workflow to modify.
    masks:
        List or tuple of pixel mask filenames to set.
    """
    workflow = workflow.copy()
    workflow[DetectorMasks] = (
        workflow[DetectorMasks]
        .map(pd.DataFrame({PixelMaskFilename: masks}).rename_axis('mask'))
        .reduce(index='mask', func=_merge)
    )
    return workflow


def with_banks(workflow: sciline.Pipeline, banks: Iterable[str]) -> sciline.Pipeline:
    """
    Return modified workflow with bank names set.

    Parameters
    ----------
    workflow:
        Workflow to modify.
    banks:
        List or tuple of bank names to set.
    """
    workflow = workflow.copy()
    banks = pd.DataFrame({NeXusDetectorName: banks}).rename_axis('bank')
    for run in (SampleRun, BackgroundRun):
        for part in (Numerator, Denominator):
            workflow[CleanSummedQ[run, part]] = (
                workflow[CleanSummedQ[run, part]]
                .map(banks)
                .reduce(index='bank', func=merge_contributions)
            )
    return workflow


def _set_runs(
    pipeline: sciline.Pipeline, runs: Iterable[str], key: Hashable, axis_name: str
) -> sciline.Pipeline:
    pipeline = pipeline.copy()
    runs = pd.DataFrame({Filename[key]: runs}).rename_axis(axis_name)
    for part in (Numerator, Denominator):
        pipeline[CleanSummedQ[key, part]] = (
            pipeline[CleanSummedQ[key, part]]
            .map(runs)
            .reduce(index=axis_name, func=merge_contributions)
        )
    return pipeline


def with_sample_runs(
    workflow: sciline.Pipeline, runs: Iterable[str]
) -> sciline.Pipeline:
    """
    Return modified workflow with sample run filenames set.

    Parameters
    ----------
    workflow:
        Workflow to modify.
    runs:
        List or tuple of sample run filenames to set.
    """
    return _set_runs(workflow, runs, SampleRun, 'sample_run')


def with_background_runs(
    workflow: sciline.Pipeline, runs: Iterable[str]
) -> sciline.Pipeline:
    """
    Return modified workflow with background run filenames set.

    Parameters
    ----------
    workflow:
        Workflow to modify.
    runs:
        List or tuple of background run filenames to set.
    """
    return _set_runs(workflow, runs, BackgroundRun, 'background_run')
